#!/bin/sh
# FILE: oommf.tcl
#
# This Tcl script is the master application of the OOMMF project.
# It must be evaluted by a Tcl interpreter such as tclsh or wish
# (or whatever other name they are known by on your platform).  
#
# On Unix platforms, it may be necessary to edit line 17 of this file
# so that the second word is the name of a Tcl interpreter shell 
# installed on your computer. 
#
# On Windows platforms, when Tcl/Tk was installed, an association was
# placed in the registry to recognize the file extension .tcl of this
# file and use the installed wish program to evaluate it.
#
#    v--Edit here if necessary \
exec tclsh "$0" ${1+"$@"}
#
# Usage:
# -----
# oommf.tcl			;# Launches $oommf(defaultArgv)
# oommf.tcl foo			;# Launches latest version of foo
# oommf.tcl foo 3		;# Launches latest version of foo compatible 
#				;# with 3
# oommf.tcl foo 3 arg1 arg2	;# Launches latest version of foo compatible 
#				;# with 3 and passes it arg1 and arg2
# oommf.tcl foo arg1 arg2	;# Launches latest version of foo and passes
#				;# it arg1 and arg2

# This whole script is wrapped in a 'catch' so we can process any errors
# in such a way that the error message is displayed to the user regardless
# of platform and Tcl version
if {[catch {

# DOME: Find some way to update these values automatically
set oommf(version) 1.0

# Set default values
set oommf(defaultArgv) mmlaunch

# Robust determination of the absolute, direct pathname of this script:
set cwd [pwd]
set myAbsoluteName [file join $cwd [info script]]
if {[catch {cd [file dirname $myAbsoluteName]} msg]} {
    error "" \
"Can't determine direct pathname for $myAbsoluteName:
	$msg"
}
set myAbsoluteName [file join [pwd] [file tail $myAbsoluteName]]
cd $cwd

# Create search path for OOMMF runtime library
set oommf(libPath) {}
if {[info exists env(OOMMF_LIBRARY)] \
        && [string match absolute [file pathtype $env(OOMMF_LIBRARY)]] \
        && [file isdirectory [file join $env(OOMMF_LIBRARY) app]]} {
    lappend oommf(libPath) $env(OOMMF_LIBRARY)
}
set _ [file join [file dirname [file dirname $myAbsoluteName]] lib oommf]
if {[file isdirectory [file join $_ app]]} {
    lappend oommf(libPath) $_
}
set _ [file dirname [file dirname [file dirname $myAbsoluteName]]]
if {[file isdirectory [file join $_ app]]} {
    lappend oommf(libPath) $_
}
if {![llength $oommf(libPath)]} {
    error "" "Unable to find the OOMMF runtime library.\n\n\
Set the environment variable OOMMF_LIBRARY to its absolute location."
}

# Search relative to the directories in oommf(libPath) for a file containing
# local modifications.  Source only the first (readable) one found. 
foreach lib $oommf(libPath) {
    set local [file join $lib app oommf$oommf(version) local oommf.tcl]
    if {[file isfile $local] && [file readable $local]} {
        uplevel #0 source [list $local]
        break
    } else {
        set local [file join $lib app oommf local oommf.tcl]
        if {[file isfile $local] && [file readable $local]} {
            uplevel #0 source [list $local]
            break
        }
    }
}

# Construct search path for extensions
foreach lib $oommf(libPath) {
    set dir [file join $lib ext]
    if {[file isdirectory $dir] && [file readable $dir] \
            && [string match absolute [file pathtype $dir]]} {
        if {[lsearch -exact $auto_path $dir] < 0} {
            lappend auto_path $dir
        }
    }
}

# Locate a version of the Oc (OOMMF core) extension and set it up for loading.
#
# The catch is here to protect against errors in the pkgIndex.tcl files
# of other packages which are not caught in pre-8.0 versions of Tcl.
if {[catch {package require Oc}]} {
    # Either we couldn't find an Oc extension, or the pkgIndex.tcl file
    # of some other package contained an error and we're running a
    # pre-8.0 version of Tcl.  In the latter case, it's possible that
    # before the error was encountered, the pkgIndex.tcl file for the
    # Oc extension was evaluated.  In that case, the workings of the
    # 'package require' command have been redefined to be robust against
    # errors in pkgIndex.tcl files.  It makes sense to try again.  If it
    # still can't find it this time, either the Oc extension really is 
    # missing, or the bad pkgIndex.tcl file is evaluated before the
    # pkgIndex.tcl file for the Oc extension, and there's nothing we can do.
    package require Oc
}

# Search relative to $oommf(libPath) for applications
Oc_Application SetSearchPath $oommf(libPath)

set fork &
while {[string match -* [lindex $argv 0]]} {
    regexp -- -(.*) [lindex $argv 0] _ option
    set argv [lrange $argv 1 end]
    switch -exact $option {
        nofork {
            set fork ""
        }
        - {
            break
        }
        default {
            error "Unrecognized option: -$option"
        }
    }
}

# Use default if no command line
if {[llength $argv] == 0} {
    set argv $oommf(defaultArgv)
}

if {[catch {
    if {[llength $argv] == 0} {
        error "No default application"
    } elseif {[regexp {^[0-9.]+$} [lindex $argv 1]]} {
        eval Oc_Application Exec [list [lrange $argv 0 1]] \
                [lrange $argv 2 end] $fork
    } else {
        eval Oc_Application Exec $argv $fork
    }
} msg]} {
    Oc_Log Log $msg error
    exit 1
} elseif {[string match {} $fork]} {
    puts $msg
}

} catchmsg]} {
    # The wish shell didn't portably display errors in startup scripts to 
    # the user until Tk 8.0.  If an earlier version of Tk is evaluating 
    # this script, we have to display the message ourselves.  The 'tk_dialog'
    # command is most reliable since it's been around the longest.
    if {[info exists tk_version] && $tk_version < 8} {
        tk_dialog .error "Error in startup script" $errorInfo error 0 OK
        exit 1
    } else {
        return -code error -errorinfo $errorInfo -errorcode $errorCode ""
    }
}

# In case this script is sourced by a 'wish' shell, explicitly call 'exit'
# so we don't enter the event loop
exit 0

