/* FILE: vector.h                 -*-Mode: c++-*-
 *
 * 3D vector C++ template declarations, and associated structures,
 * for use with both REAL4 and REAL8 data types.
 *
 */

#ifndef _NB_VECTOR
#define _NB_VECTOR

#include "evoc.h"

/* End includes */     /* Optional directive to build.tcl */

/////////////////// 3D Vector Template Declaration ////////////////////
// NOTE: "T" should be either REAL4 or REAL8.
template<class T> class Nb_Vec3 {
public:
  T x,y,z;  // Components
  void Set(T nx,T ny,T nz) { x=nx; y=ny; z=nz; }
  Nb_Vec3(void) { x=y=z=0.0; } // Default constructor
  Nb_Vec3(T nx,T ny,T nz) { Set(nx,ny,nz); }
  Nb_Vec3(const Nb_Vec3 &v) : x(v.x), y(v.y), z(v.z) {} // Copy constructor
  // Assignment operator
  Nb_Vec3<T>& operator=(const Nb_Vec3<T> &v) {
    x=v.x; y=v.y; z=v.z; return *this;
  }
  BOOL Equals(const Nb_Vec3<T> &v) const {
    return (x==v.x  && y==v.y && z==v.z);
  }
  BOOL Bigger(const Nb_Vec3<T> &v) const {
    return (x>=v.x  && y>=v.y && z>=v.z);
  }
  T MagSq() const { return x*x+y*y+z*z; }
  T Mag() const { return sqrt(MagSq()); }
  Nb_Vec3<T> & operator+=(const Nb_Vec3<T>& v);
  Nb_Vec3<T> & operator-=(const Nb_Vec3<T>& v);
  Nb_Vec3<T> & operator*=(T a);
  friend Nb_Vec3<T> operator+(const Nb_Vec3<T>& v,const Nb_Vec3<T>& w);
  friend Nb_Vec3<T> operator-(const Nb_Vec3<T>& v,const Nb_Vec3<T>& w);
  friend Nb_Vec3<T> operator*(T a,const Nb_Vec3<T>& v);
  friend Nb_Vec3<T> operator*(const Nb_Vec3<T>& v,T a);
};

template<class T> Nb_Vec3<T> & Nb_Vec3<T>::operator+=(const Nb_Vec3<T>& v)
{
  x+=v.x; y+=v.y; z+=v.z; return *this;
}

template<class T> Nb_Vec3<T> & Nb_Vec3<T>::operator-=(const Nb_Vec3<T>& v)
{
  x-=v.x; y-=v.y; z-=v.z; return *this;
}

template<class T> Nb_Vec3<T> & Nb_Vec3<T>::operator*=(T a)
{
  x*=a; y*=a; z*=a; return *this;
}

template<class T> Nb_Vec3<T> operator+(const Nb_Vec3<T>& v,
				       const Nb_Vec3<T>& w)
{
  Nb_Vec3<T> temp;
  temp.x=v.x+w.x;  temp.y=v.y+w.y;  temp.z=v.z+w.z;
  return temp;
}

template<class T> Nb_Vec3<T> operator-(const Nb_Vec3<T>& v,
				       const Nb_Vec3<T>& w)
{
  Nb_Vec3<T> temp;
  temp.x=v.x-w.x;  temp.y=v.y-w.y;  temp.z=v.z-w.z;
  return temp;
}

template<class T> Nb_Vec3<T> operator*(T a,const Nb_Vec3<T>& v)
{
  Nb_Vec3<T> temp;
  temp.x=a*v.x;  temp.y=a*v.y;  temp.z=a*v.z;
  return temp;
}

template<class T> inline Nb_Vec3<T> operator*(const Nb_Vec3<T>& v,T a)
{
  return a*v;
}

//////////////////////////////////////////////////////////////////////////
// Position _and_ Vector structure
template<class T> struct Nb_LocatedVector {
public:
  Nb_Vec3<T> location;
  Nb_Vec3<T> value;
  Nb_LocatedVector() { location.Set(0.,0.,0.); value.Set(0.,0.,0.); }
  Nb_LocatedVector(const Nb_Vec3<T> &newloc,const Nb_Vec3<T> &newval) {
    location=newloc; value=newval;
  }
};

//////////////////////////////////////////////////////////////////////////
// Bounding Box
template<class T> class Nb_BoundingBox
{ // Note: Empty boxes are well-defined.  See Empty() function below.
  Nb_Vec3<T> bbmin;
  Nb_Vec3<T> bbmax;
public:
  BOOL IsEmpty() const { 
    return (bbmin.x>bbmax.x || bbmin.y>bbmax.y || bbmin.z>bbmax.z);
  }
  void Reset() { bbmin.Set(0.,0.,0.); bbmax.Set(-1.,-1.,-1.); }
  void SortAndSet(const Nb_Vec3<T> &corner1,
		  const Nb_Vec3<T> &corner2);
  void Set(const Nb_Vec3<T> &corner1,const Nb_Vec3<T> &corner2);
  void Set(const Nb_BoundingBox<T> &_other) {
    Set(_other.bbmin,_other.bbmax);
  }
  BOOL ExpandWith(const Nb_Vec3<T> &include_point); // Expands box as
  /// necessary to include 'include_point'.  Returns 1 if expansion occurred,
  /// 0 otherwise.
  BOOL ExpandWith(const Nb_BoundingBox<T> &other_box); // Expands box as
  /// necessary to include 'other_box'.  Returns 1 if expansion occurred,
  /// 0 otherwise
  friend Nb_BoundingBox<T>
  IntersectBoundingBoxes(const Nb_BoundingBox<T>& box1,
			 const Nb_BoundingBox<T>& box2);
  /// Returns the box resulting from the intersection of box1 and box2.
  /// If the two imports have empty intersection, then the returned
  /// box has .bbmin=(0,0,0),.bbmax=(-1,-1,-1)
  
  // Default constructor
  Nb_BoundingBox(void) { Reset(); }
  // Copy constructor
  Nb_BoundingBox(const Nb_BoundingBox<T>& newbb) :
    bbmin(newbb.bbmin), bbmax(newbb.bbmax) {}

  // Construction from two Nb_Vec3<T>'s.
  Nb_BoundingBox(const Nb_Vec3<T>& v1,
                 const Nb_Vec3<T> v2) : bbmin(v1), bbmax(v2) {}

  // Assignment operator
  Nb_BoundingBox<T>& operator=(const Nb_BoundingBox<T>& bb) { 
    bbmin=bb.bbmin; bbmax=bb.bbmax; return *this;
  }
  BOOL IsIn(const Nb_Vec3<T> &location) const;
  // Returns 1 if location is inside the box.

  BOOL MoveInto(Nb_Vec3<T> &location) const;
  // Adjusts location minimally to be inside bbox.
  // Returns 1 if location was already inside bbox (no shift done).
  // ASSUMES bbox is non-empty!

  void GetMinPt(Nb_Vec3<T> &minpt) const { minpt=bbmin; }
  void GetMaxPt(Nb_Vec3<T> &maxpt) const { maxpt=bbmax; }
  void GetExtremes(Nb_Vec3<T> &minpt,Nb_Vec3<T> &maxpt) const {
    minpt=bbmin; maxpt=bbmax;
  }
  T GetWidth(void)  const { return OC_MAX(bbmax.x-bbmin.x,0.); }
  T GetHeight(void) const { return OC_MAX(bbmax.y-bbmin.y,0.); }
  T GetDepth(void)  const { return OC_MAX(bbmax.z-bbmin.z,0.); }
  T GetVolume(void) const { return GetWidth()*GetHeight()*GetDepth(); }
  Nb_Vec3<T> GetDiag(void) const; // Returns max-min if box is not empty,
  /// 0 otherwise.
};

template<class T> BOOL
Nb_BoundingBox<T>::IsIn(const Nb_Vec3<T> &location) const
{
  return (location.Bigger(bbmin) && (bbmax.Bigger(location)));
}

template<class T> BOOL
Nb_BoundingBox<T>::MoveInto(Nb_Vec3<T> &location) const
{ // Adjusts location minimally to be inside bbox.
  // Returns 1 if location was already inside bbox (no shift done).
  // ASSUMES bbox is non-empty!
  BOOL is_in=1;
  if(location.x<bbmin.x) {
    location.x=bbmin.x; is_in=0;
  } else if(location.x>bbmax.x) {
    location.x=bbmax.x; is_in=0;
  }
  if(location.y<bbmin.y) {
    location.y=bbmin.y; is_in=0;
  } else if(location.y>bbmax.y) {
    location.y=bbmax.y; is_in=0;
  }
  if(location.z<bbmin.z) {
    location.z=bbmin.z; is_in=0;
  } else if(location.z>bbmax.z) {
    location.z=bbmax.z; is_in=0;
  }
  return is_in;
}

template<class T> void
Nb_BoundingBox<T>::SortAndSet(const Nb_Vec3<T> &corner1,
                              const Nb_Vec3<T> &corner2) {
  if(corner1.x<corner2.x) { bbmin.x=corner1.x; bbmax.x=corner2.x; }
  else                    { bbmin.x=corner2.x; bbmax.x=corner1.x; }
  if(corner1.y<corner2.y) { bbmin.y=corner1.y; bbmax.y=corner2.y; }
  else                    { bbmin.y=corner2.y; bbmax.y=corner1.y; }
  if(corner1.z<corner2.z) { bbmin.z=corner1.z; bbmax.z=corner2.z; }
  else                    { bbmin.z=corner2.z; bbmax.z=corner1.z; }
}

template<class T> void
Nb_BoundingBox<T>::Set(const Nb_Vec3<T> &corner1,
                       const Nb_Vec3<T> &corner2) {
  bbmin.x=corner1.x; bbmax.x=corner2.x;
  bbmin.y=corner1.y; bbmax.y=corner2.y;
  bbmin.z=corner1.z; bbmax.z=corner2.z;
}

template<class T> BOOL
Nb_BoundingBox<T>::ExpandWith(const Nb_Vec3<T> &include_point)
{ // Expands box as necessary to include 'include_point'. Returns
  // 1 if expansion occurred, 0 otherwise.
  if(IsEmpty()) {
    bbmin=bbmax=include_point;
    return 1;
  }
  BOOL expand=0;
  if(include_point.x<bbmin.x) { bbmin.x=include_point.x; expand=1; }
  if(include_point.y<bbmin.y) { bbmin.y=include_point.y; expand=1; }
  if(include_point.z<bbmin.z) { bbmin.z=include_point.z; expand=1; }
  if(include_point.x>bbmax.x) { bbmax.x=include_point.x; expand=1; }
  if(include_point.y>bbmax.y) { bbmax.y=include_point.y; expand=1; }
  if(include_point.z>bbmax.z) { bbmax.z=include_point.z; expand=1; }
  return expand;
}

template<class T> BOOL
Nb_BoundingBox<T>::ExpandWith(const Nb_BoundingBox<T>& other_box)
{ // Expands box as necessary to include 'other_box'.  Returns 1 if
  // expansion occurred, 0 otherwise
  if(IsEmpty()) {
    if(other_box.IsEmpty()) return 0;
    Set(other_box);
    return 1;
  }
  BOOL expand=0;
  if(other_box.bbmin.x<bbmin.x) { bbmin.x=other_box.bbmin.x; expand=1; }
  if(other_box.bbmin.y<bbmin.y) { bbmin.y=other_box.bbmin.y; expand=1; }
  if(other_box.bbmin.z<bbmin.z) { bbmin.z=other_box.bbmin.z; expand=1; }
  if(other_box.bbmax.x>bbmax.x) { bbmax.x=other_box.bbmax.x; expand=1; }
  if(other_box.bbmax.y>bbmax.y) { bbmax.y=other_box.bbmax.y; expand=1; }
  if(other_box.bbmax.z>bbmax.z) { bbmax.z=other_box.bbmax.z; expand=1; }
  return expand;
}

template<class T> Nb_BoundingBox<T>
IntersectBoundingBoxes(const Nb_BoundingBox<T>& box1,
                       const Nb_BoundingBox<T>& box2)
{ // Returns the box resulting from the intersection of box1 and box2.
  // If the two imports have empty intersection, then the returned
  // box has .bbmin=(0,0,0), .bbmax=(-1,-1,-1)
  Nb_BoundingBox<T> box;
  box.Reset();  // Safety
  if(box1.IsEmpty() || box2.IsEmpty()) return box;
  Nb_Vec3<T> bbmin,bbmax;
  bbmin.x=OC_MAX(box1.bbmin.x,box2.bbmin.x);
  bbmin.y=OC_MAX(box1.bbmin.y,box2.bbmin.y);
  bbmin.z=OC_MAX(box1.bbmin.z,box2.bbmin.z);
  bbmax.x=OC_MIN(box1.bbmax.x,box2.bbmax.x);
  bbmax.y=OC_MIN(box1.bbmax.y,box2.bbmax.y);
  bbmax.z=OC_MIN(box1.bbmax.z,box2.bbmax.z);
  box.Set(bbmin,bbmax);
  if(box.IsEmpty()) box.Reset();
  return box;
}

template<class T> Nb_Vec3<T> Nb_BoundingBox<T>::GetDiag(void) const { 
  if(IsEmpty()) return Nb_Vec3<T>(0.,0.,0.);
  return bbmax-bbmin;
}

#endif // _NB_VECTOR
